import os
import re
import pymongo
import json
from math import ceil
from typing import List
from tweet_manager.lib import json2csv, format_csv

# Logger
import logging
logger = logging.getLogger(__name__)


def write_tweets_to_files(
        host: str,
        port: int,
        database: str,
        pagesize: int,
        header: str,
        output_dir: str,
        page_index: List[int])\
        -> None:
    logger.info("Hi there! write_tweets_to_files executing")
    client = pymongo.MongoClient(host, port)
    database_tweets = client[database]["tweets"]
    tweets_page = get_tweets_page(database_tweets, pagesize, page_index)
    buffer_tweets = {}
    for tweet in tweets_page:
        # Get output path and contents for the new CSV file
        csv_tweet_output_path =\
            create_tweet_output_path(header, tweet, output_dir)
        csv_tweet_contents =\
            "\n" + str(convert_tweet_to_csv(header, tweet))

        # Check if buffer exists for the file. If not, add to dictionary
        if csv_tweet_output_path not in buffer_tweets.keys():
            buffer_tweets[csv_tweet_output_path] = ["", 0]

        # Update the buffer adding the tweet and increasing tweet count
        buffer_tweets[csv_tweet_output_path][0] += csv_tweet_contents
        # buffer_tweets[csv_tweet_output_path][1] += 1

    client.close()
    # Perform the write operations in each of the files
    for output_path in buffer_tweets.keys():
        with open(output_path, "a") as tweet_writer:
            tweet_writer.write(buffer_tweets[output_path][0])


def get_tweet_output_path(tweet: dict, output_dir: str) -> str:
    """
    Returns same output than create_tweet_output_path, but without
    touching the filesystem
    """

    collection_path = output_dir

    # Extract year, month and date from the tweet using a regex
    matchObj = re.search(
        r"^(\d{4})-(\d{2})-(\d{2}) \d{2}:\d{2}:\d{2}$",
        str(tweet["created_at"]))
    year = matchObj.group(1)
    month = matchObj.group(2)
    day = matchObj.group(3)

    tweet_output_file = os.path.join(
        collection_path, year, month, day + ".csv")

    return tweet_output_file


def create_tweet_output_path(
        header: str,
        tweet: dict,
        output_dir: str)\
        -> str:
    try:
        collection_path = create_task_database_structure(output_dir)
    except FileExistsError as e:
        collection_path = e.filename

    # Extract year, month and date from the tweet using a regex
    matchObj = re.search(
        r"^(\d{4})-(\d{2})-(\d{2}) \d{2}:\d{2}:\d{2}$",
        str(tweet["created_at"]))
    year = matchObj.group(1)
    month = matchObj.group(2)
    day = matchObj.group(3)
    date = (year, month, "")

    # Classify the tweet chronologically
    tweet_output_path = json2csv.mkdir_tweet_date(date, collection_path)
    tweet_output_file = os.path.join(tweet_output_path, day + ".csv")

    # If the CSV file didn't already exist, initialize it with a header
    if os.path.isfile(tweet_output_file) is False:
        with open(tweet_output_file, "w") as fw:
            fw.write(header.strip())

    return tweet_output_file


def create_task_database_structure(
        output_dir: str)\
        -> str:
    """
    Generate the following directory tree: a top dir that will contain
    all the tweet collections if it didn't exist yet and within it the top
    directory for this task
    """

    # Create the root directory for the tweet collection
    (output_dir, db_name) = os.path.split(output_dir)
    if not os.path.isdir(output_dir):
        logger.info(
            "Building directory to contain the collected tweets at: "
            + os.path.abspath(output_dir)
        )
        os.mkdir(output_dir)
    collection_path = os.path.join(output_dir, db_name)
    if not os.path.isdir(collection_path):
        logger.info("Initializing collection " + db_name + "...")
        os.mkdir(collection_path)

    return collection_path


def convert_tweet_to_csv(header: str, tweet: dict) -> str:
    # Flatten the tweet and store it in status_flat
    status_flat = json2csv.flatten_dictionary(tweet)

    # Convert the flat JSON to CSV format
    # 1st arg: flat tweet, 2nd arg: activate array compression, 3rd arg:
    # number of array compression levels, 4th arg: remove dollars mode
    status_csv = json2csv.json2csv(status_flat, True, 5, False)

    csv_appendable_line = format_csv.get_csv_line(header, status_csv)

    return csv_appendable_line


#########################
#  TWEET DB PAGINATION  #
#########################
def get_page_index(
        collection: pymongo.collection.Collection,
        page_size: int)\
        -> List[int]:
    """
    Get an iterator with ints between 0 and N-1, where N is the number
    of pages of the collection for the given page size
    """

    return list(range(0, ceil(collection.count() / page_size)))


def get_tweets_page(
        collection: pymongo.collection.Collection,
        page_size: int,
        num_page: int)\
        -> pymongo.cursor.Cursor:
    """
    Returns a pymongo cursor pointing to the MongoDB entries comprised
    in the current page
    """

    tweets = collection.find().skip(num_page * page_size).limit(page_size)
    return tweets


#########################
#  METADATA GENERATION  #
#########################
def generate_metadata_file(collection_path: str) -> None:
    """
    Once all the CSV files have been created, generate a metadata file
    with information about the number of tweets in each of the CSVs by
    making a simple count of lines
    """

    metadata = {}
    metadata["files"] = {}
    for root, dirs, files in os.walk(collection_path):
        for f in files:
            file_path = os.path.join(root, f)
            relative_path = os.path.relpath(file_path, collection_path)
            metadata["files"][relative_path] = {}
            metadata["files"][relative_path]["count"] =\
                file_length(file_path)
    output_path = os.path.join(collection_path, ".metadata.json")
    with open(output_path, 'w') as f:
        json.dump(metadata, f)


def file_length(file_path: str) -> int:
    """
    Calculate number of lines of a file
    """

    with open(file_path) as f:
        for i, l in enumerate(f):
            pass
        return i
